import unittest

import numpy as np
import tensorflow as tf

from tensorflow.python.client import device_lib
from tensorflow.python.ops import gradient_checker
from tensorflow.python.framework import constant_op

from qrnn import fo_pool, time_major_fo_pool_unsliced, batch_major_fo_pool_unsliced

def np_fo_pooling(x, forget, initial_state, time_major):
    if not time_major:
        return np.transpose(np_fo_pooling(np.transpose(x, (1, 0, 2)),
                                          np.transpose(forget, (1, 0, 2)),
                                          initial_state,
                                          time_major=True), (1, 0, 2))
    timesteps, batch, hidden = x.shape
    dst = np.zeros((timesteps + 1, batch, hidden), dtype=x.dtype)
    dst[0] = initial_state
    for ts in range(1, timesteps + 1):
        dst[ts] = (forget[ts - 1]         * x[ts - 1] +
                   (1.0 - forget[ts - 1]) * dst[ts - 1])
    return dst[1:]


class TestFoPool(unittest.TestCase):
    """ Tests the FoPool operator """

    def setUp(self):
        # Obtain a list of GPU device specifications ['/gpu:0', '/gpu:1', ...]
        self.gpu_devs = [d.name for d in device_lib.list_local_devices()
                         if d.device_type == 'GPU']

    def test_fo_pool(self):
        """ Test the FoPool operator """
        # List of type constraint for testing this operator
        type_permutations = [np.float32, np.float64]

        # Run test with the type combinations above
        for FT in type_permutations:
            for time_major in [False]:
                self._impl_test_fo_pool(FT, time_major)

    def _impl_test_fo_pool(self, FT, time_major):
        """ Implementation of the FoPool operator test """
        # Create input variables
        timesteps = 20
        batch_size = 32
        channels = 64
        if time_major:
            shape = (timesteps, batch_size, channels)
        else:
            shape = (batch_size, timesteps, channels)
        x = np.random.random(size=shape).astype(FT)
        forget = np.random.uniform(0, 1, size=shape).astype(FT)
        initial_state = np.random.random(size=(batch_size, channels)).astype(FT)

        # Argument list
        np_args = [x, forget, initial_state]
        # Argument string name list
        arg_names = ["x", "forget", "initial_state"]
        # Constructor tensorflow variables
        tf_args = [tf.Variable(v, name=n) for v, n in zip(np_args, arg_names)]

        def _pin_op(device, *tf_args):
            """ Pin operation to device """
            with tf.device(device):
                return fo_pool(*tf_args, time_major=time_major)

        # Pin operation to CPU
        cpu_op = _pin_op("/cpu:0", *tf_args)

        # Run the op on all GPUs
        gpu_ops = [_pin_op(d, *tf_args) for d in self.gpu_devs]

        # Initialise variables
        init_op = tf.global_variables_initializer()

        with tf.Session() as S:
            S.run(init_op)
            cpu_result = S.run(cpu_op)
            self.assertEqual(cpu_result.shape, shape)
            gpu_results = S.run(gpu_ops)
            for gpu_result in gpu_results:
                self.assertEqual(gpu_result.shape, shape)
            expected = np_fo_pooling(x, forget, initial_state, time_major=time_major)
            self.assertTrue(np.allclose(cpu_result, expected))
            for gpu_result in gpu_results:
                self.assertTrue(np.allclose(gpu_result, expected))

    def test_time_major_fo_pool_grad(self):
        """ Test the FoPool Gradient operator """
        # List of type constraint for testing this operator
        type_permutations = [(np.float32, 1e-2), (np.float64, 1e-4)]

        # Run test with the type combinations above
        for FT, tolerance in type_permutations:
            self._impl_test_time_major_fo_pool_grad(FT, tolerance)


    def _impl_test_time_major_fo_pool_grad(self, FT, tolerance):
        shape = (5, 3, 2)
        np_args = [np.random.random(size=shape).astype(FT),
                   np.random.uniform(0, 1, size=shape).astype(FT),
                   np.random.random(size=shape[1:]).astype(FT)]
        with tf.Session() as S:
            tf_args = [constant_op.constant(arg, shape=arg.shape, dtype=FT) for arg in np_args]
            y = tf.reduce_sum(time_major_fo_pool_unsliced(*tf_args))
            for d in ["cpu"] + self.gpu_devs:
                with tf.device(d):
                    err = gradient_checker.compute_gradient_error(
                      tf_args, [arg.shape for arg in np_args], y, [],
                      x_init_value=np_args)
                self.assertLess(err, tolerance)

    def test_batch_major_fo_pool_grad(self):
        """ Test the FoPool Gradient operator """
        # List of type constraint for testing this operator
        type_permutations = [(np.float32, 1e-2), (np.float64, 1e-4)]

        # Run test with the type combinations above
        for FT, tolerance in type_permutations:
            self._impl_test_batch_major_fo_pool_grad(FT, tolerance)


    def _impl_test_batch_major_fo_pool_grad(self, FT, tolerance):
        shape = (3, 5, 2)
        np_args = [np.random.random(size=shape).astype(FT),
                   np.random.uniform(0, 1, size=shape).astype(FT),
                   np.random.random(size=(shape[0], shape[-1])).astype(FT)]
        with tf.Session() as S:
            tf_args = [constant_op.constant(arg, shape=arg.shape, dtype=FT) for arg in np_args]
            y = tf.reduce_sum(batch_major_fo_pool_unsliced(*tf_args))
            for d in ["cpu"] + self.gpu_devs:
                with tf.device(d):
                    err = gradient_checker.compute_gradient_error(
                      tf_args, [arg.shape for arg in np_args], y, [],
                      x_init_value=np_args)
                self.assertLess(err, tolerance)


if __name__ == "__main__":
    unittest.main()



"""
Quasi-Recurrent Neural Network (QRNN) for Tensorflow
----------------------------------------------------

This repository contains a Tensorflow implementation of
[Salesforce Research](https://einstein.ai/)'s
[Quasi-Recurrent Neural Networks](https://arxiv.org/abs/1611.01576)
paper. It supports batch-major or time-major inputs in
single or double precision.

From the authors:
> The QRNN provides similar accuracy to the LSTM but can be betwen
> 2 and 17 times faster than the highly optimized NVIDIA cuDNN
> LSTM implementation depending on the use case.

If you use this code or their results in your research, you should cite:

@article{bradbury2016quasi,
  title={{Quasi-Recurrent Neural Networks}},
  author={Bradbury, James and Merity, Stephen and Xiong, Caiming and Socher, Richard},
  journal={International Conference on Learning Representations (ICLR 2017)},
  year={2017}
}

Usage
-----

Use QRNNs as you would use LSTMs or RNNs, to
encode order-specific information:

```
import qrnn

# input sequence in Batch, Time, Channels format:
inputs = tf.placeholder(tf.float32, [None, None, 128])
encoded = qrnn.qrnn(inputs)

with tf.Session() as sess:
  sess.run(tf.global_variables_initializer())
  out = sess.run(encoded, {inputs: my_data})
```

"""
import tensorflow as tf
from os.path import join, dirname, realpath

SCRIPT_DIR = dirname(realpath(__file__))

def get_ext_filename(ext_name):
    from distutils.sysconfig import get_config_var
    ext_path = ext_name.split('.')
    ext_suffix = get_config_var('EXT_SUFFIX')
    return join(*ext_path) + ext_suffix


qrnn_lib = tf.load_op_library(join(SCRIPT_DIR, get_ext_filename("qrnn_lib")))

time_major_fo_pool_unsliced = qrnn_lib.time_major_fo_pool
time_major_bwd_fo_pool = qrnn_lib.time_major_bwd_fo_pool

batch_major_fo_pool_unsliced = qrnn_lib.batch_major_fo_pool
batch_major_bwd_fo_pool = qrnn_lib.batch_major_bwd_fo_pool

@tf.RegisterGradient("TimeMajorFoPool")
def _fo_pool_grad(op, grad):
    return time_major_bwd_fo_pool(h=op.outputs[0], x=op.inputs[0],
                                  forget=op.inputs[1], gh=grad)

@tf.RegisterGradient("BatchMajorFoPool")
def _fo_pool_grad(op, grad):
    return batch_major_bwd_fo_pool(h=op.outputs[0], x=op.inputs[0],
                                   forget=op.inputs[1], gh=grad)


def fo_pool(x, forget, initial_state=None, time_major=False):
    """Applies a single layer Quasi-Recurrent Neural Network (QRNN) to an input sequence.
    Args:
        x: Tensor, input values in [Batch, Time, Channels] format,
           float32 or double
           or [Time, Batch, Channels] if time_major
        forget: Tensor, input values in [Batch, Time, Channels] format,
           float32 or double. Usually in the range 0-1.
           or [Time, Batch, Channels] if time_major
        initial_state: Tensor, initial hidden state values in [Batch, Channels] format,
           float32 or double.

    Returns:
        Tensor: fo_pooled output, [Batch, Time, Channels] format
                or [Time, Batch, Channels] if time_major
    """
    if initial_state is None:
        initial_state = tf.zeros((tf.shape(x)[1] if time_major else tf.shape(x)[0],
                                  tf.shape(x)[2]), dtype=x.dtype)
    if time_major:
        return time_major_fo_pool_unsliced(x, forget, initial_state)[1:]
    else:
        return batch_major_fo_pool_unsliced(x, forget, initial_state)[:, 1:]


def qrnn(inputs, num_outputs, window=2, output_gate=True,
         activation_fn=tf.tanh, gate_activation_fn=tf.nn.sigmoid,
         padding="SAME", initial_state=None, time_major=False, scope=None,
         **kwargs):
    """Applies a single layer Quasi-Recurrent Neural Network (QRNN) to an input sequence.
    Args:
        inputs: Tensor, input values in [Batch, Time, Channels] format,
                float32 or double, or [Time, Batch, Channels] if time_major
        window: int, number of values each gating depends on (default=2).
        num_outputs: int, Number of output channels
        keep_prob: float, zoneout dropout probability
        is_training: bool, whether to apply dropout mask
        output_gate: bool, use a gating mechanism on the output
        activation_fn: function, default tanh
        gate_activation_fn: function, default sigmoid
        padding: str, SAME or VALID.
        initial_state: Tensor/None, optional, initializes the QRNN to that value.
        time_major: bool, whether inputs have time-dimension first or second.
        scope: str/None, what to prefix the name the variables under this layer.

    Returns:
        Tensor : qrnn_output, [Batch, Time, Channels] or
                [Time, Batch, Channels] if time_major
    """
    with tf.variable_scope(scope or "QRNNLayer"):
        conv1d_channels = 3 * num_outputs if output_gate else 2 * num_outputs
        if time_major:
            # go to batch_major for convolution if needed
            inputs_batch_major = tf.transpose(inputs, (1, 0, 2), name="InputsBatchMajor")
        else:
            inputs_batch_major = inputs
        gate_values = tf.layers.conv1d(inputs_batch_major,
                                       filters=conv1d_channels,
                                       kernel_size=window,
                                       strides=1,
                                       data_format="channels_last",
                                       name="QRNNConv1D",
                                       padding=padding,
                                       **kwargs)
        if time_major:
            # return to time_major if needed
            gate_values = tf.transpose(gate_values, (1, 0, 2),
                                       name="GateValuesTimeMajor")

        gate_values = tf.split(gate_values, 3 if output_gate else 2, axis=2)
        if output_gate:
            x, forget, output = gate_values
        else:
            x, forget = gate_values

        with tf.name_scope("GateActivations"):
            if activation_fn is not None:
                x = activation_fn(x)
            if gate_activation_fn is not None:
                forget = gate_activation_fn(forget)

        with tf.name_scope("FoPool"):
            c = fo_pool(x, forget, initial_state=initial_state,
                        time_major=time_major)

        with tf.name_scope("OutputGate"):
            h = gate_activation_fn(c) if output_gate else c
    return h, c



import  os
import sys
from os.path import join, dirname, realpath, relpath, splitext, abspath, exists, getmtime, relpath, lexists, islink
from os import walk, sep, remove, listdir, stat, symlink, pathsep
from setuptools import setup
from distutils.extension import Extension
from distutils.command.build_ext import build_ext
from setuptools.command.develop import develop
from setuptools.command.install import install
import warnings
import tempfile
import subprocess

import numpy as np
import tensorflow as tf

def find_in_path(name, path):
    "Find a file in a search path"
    #adapted fom http://code.activestate.com/recipes/52224-find-a-file-given-a-search-path/
    solutions = []
    for dir in path.split(pathsep):
        binpath = join(dir, name)
        if exists(binpath):
            solutions.append(abspath(binpath))
    if len(solutions) >= 1:
        if any("usr" in sol for sol in solutions):
            solutions = [sol for sol in solutions if "usr" in sol]
        return solutions[0]
    return None


def locate_cuda():
    """Locate the CUDA environment on the system
    Returns a dict with keys 'home', 'nvcc', 'include', and 'lib64'
    and values giving the absolute path to each directory.
    Starts by looking for the CUDAHOME env variable. If not found, everything
    is based on finding 'nvcc' in the PATH.
    """

    # first check if the CUDAHOME env variable is in use
    if "CUDAHOME" in os.environ:
        home = os.environ["CUDAHOME"]
        nvcc = join(home, "bin", "nvcc")
    else:
        # otherwise, search the PATH for NVCC
        nvcc = find_in_path("nvcc", os.environ["PATH"])
        print()
        if nvcc is None:
            return None
        home = dirname(dirname(nvcc))
        print(home)
    cudaconfig = {"nvcc": nvcc,
                  "include": [join(home, "include"), join(home, "include", "cuda")],
                  "lib64": [join(home, "lib64"), join(home, "lib")]}
    for k, v in cudaconfig.items():
        if isinstance(v, str):
            v = [v]
        all_missing = all(not exists(path) for path in v)
        if all_missing:
            raise EnvironmentError("The CUDA %s path could not be located in %r" % (k, v))

    return cudaconfig


def customize_compiler_for_nvcc(self):
    """inject deep into distutils to customize how the dispatch
    to gcc/nvcc works.
    If you subclass UnixCCompiler, it's not trivial to get your subclass
    injected in, and still have the right customizations (i.e.
    distutils.sysconfig.customize_compiler) run on it. So instead of going
    the OO route, I have this. Note, it's kindof like a wierd functional
    subclassing going on."""

    # tell the compiler it can processes .cu
    self.src_extensions.append('.cu')

    # save references to the default compiler_so and _comple methods
    default_compiler_so = self.compiler_so
    super = self._compile

    # now redefine the _compile method. This gets executed for each
    # object but distutils doesn't have the ability to change compilers
    # based on source extension: we add it.
    def _compile(obj, src, ext, cc_args, extra_postargs, pp_opts):
        if os.path.splitext(src)[1] == ".cu":
            # use the cuda for .cu files
            self.set_executable("compiler_so", CUDA["nvcc"])
            # use only a subset of the extra_postargs, which are 1-1 translated
            # from the extra_compile_args in the Extension class
            postargs = extra_postargs["nvcc"]
        else:
            postargs = extra_postargs["gcc"]

        super(obj, src, ext, cc_args, postargs, pp_opts)
        # reset the default compiler_so, which we might have changed for cuda
        self.compiler_so = default_compiler_so

    # inject our redefined _compile method into the class
    self._compile = _compile


def check_openmp_presence():
    source = """
        #include <omp.h>
        int main() {
        #ifdef _OPENMP
          return 0;
        #else
          breaks_on_purpose
        #endif
        }
    """
    some_dir = tempfile.mkdtemp()
    with open(join(some_dir, "program.cpp"), "wt") as fout:
        fout.write(source)
    try:
        out = subprocess.check_output(["g++", join(some_dir, "program.cpp"), "-o", join(some_dir, "program.o"), "-fopenmp"])
        supported = True
    except subprocess.CalledProcessError:
        supported = False
    finally:
        try:
            os.remove(join(some_dir, "program.cpp"))
        except FileNotFoundError:
            pass
        try:
            os.remove(join(some_dir, "program.o"))
        except FileNotFoundError:
            pass
    os.rmdir(some_dir)
    return supported



# run the customize_compiler
class custom_build_ext(build_ext):
    def build_extensions(self):
        customize_compiler_for_nvcc(self.compiler)
        build_ext.build_extensions(self)


def find_files_by_suffix(path, suffix):
    """Recursively find files with specific suffix in a directory"""
    for relative_path, dirs, files in walk(path):
        for fname in files:
            if fname.endswith(suffix):
                yield join(path, relative_path, fname)

CUDA = locate_cuda()
SRC_DIR = join(dirname(realpath(__file__)), "src")
TF_LIB = tf.sysconfig.get_lib()
TF_INCLUDE = tf.sysconfig.get_include()
TF_CUDA = tf.test.is_built_with_cuda()
HAS_OPENMP = check_openmp_presence()

if TF_CUDA and CUDA is None:
    warnings.warn("qrnn can run on gpu, but nvcc was not found in your path. "
                  "Either add it to your path, or set the $CUDAHOME variable.")

USE_CUDA = TF_CUDA and CUDA is not None


cu_sources = list(find_files_by_suffix(SRC_DIR, ".cu"))
cpp_sources = list(find_files_by_suffix(SRC_DIR, ".cpp"))


cmdclass = {}
include_dirs = [np.get_include(), TF_INCLUDE, SRC_DIR, join(SRC_DIR, "third_party")]
TF_FLAGS = ["-D_MWAITXINTRIN_H_INCLUDED", "-D_FORCE_INLINES", "-D_GLIBCXX_USE_CXX11_ABI=0"]
gcc_extra_compile_args = ["-g", "-std=c++11", "-fPIC", "-O3", "-march=native", "-mtune=native"] + TF_FLAGS

nvcc_extra_compile_args = []
extra_link_args = ["-fPIC"]
if HAS_OPENMP:
    gcc_extra_compile_args.append("-fopenmp")
    extra_link_args.append("-fopenmp")

if sys.platform == 'darwin':
    gcc_extra_compile_args.append('-stdlib=libc++')
    nvcc_extra_compile_args.append('-stdlib=libc++')
    extra_link_args.append('-stdlib=libc++')
else:
    extra_link_args.append("-shared")


if USE_CUDA:
    cmdclass["build_ext"] = custom_build_ext
    gcc_extra_compile_args.extend(["-D", "GOOGLE_CUDA"])
    include_dirs.extend(CUDA["include"])
    nvcc_extra_compile_args.extend(TF_FLAGS + ["-std=c++11", "-D", "GOOGLE_CUDA=1",
                               "-I", TF_INCLUDE,
                               "-x", "cu", "--compiler-options", "'-fPIC'",
                               "--gpu-architecture=sm_30", "-lineinfo",
                               "-Xcompiler", "-std=c++98"] + ["-I" + path for path in CUDA["include"]])
    extra_compile_args = {"gcc": gcc_extra_compile_args,
                          "nvcc": nvcc_extra_compile_args}
    runtime_library_dirs = CUDA['lib64']
else:
    cu_sources = []
    extra_compile_args = gcc_extra_compile_args
    runtime_library_dirs = []


ext = Extension("qrnn_lib",
                sources=cu_sources + cpp_sources,
                library_dirs=[TF_LIB],
                libraries=["tensorflow_framework"],
                language="c++",
                runtime_library_dirs=runtime_library_dirs,
                # this syntax is specific to this build system
                # we're only going to use certain compiler args with nvcc and not with gcc
                # the implementation of this trick is in customize_compiler() below
                extra_compile_args=extra_compile_args,
                extra_link_args=extra_link_args,
                include_dirs=include_dirs)

setup(name='qrnn',
      # random metadata. there's more you can supploy
      author="Jonathan Raiman",
      author_email="jonathanraiman@gmail.com",
      version="0.2.3",
      install_requires=["numpy", "tensorflow>=1.4"],
      ext_modules = [ext],
      py_modules=["qrnn"],
      # inject our custom trigger
      cmdclass=cmdclass,
      zip=False)
